# Copyright (c) 2013 Julien Danjou <julien@danjou.info>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from blazarnova.scheduler.filters import blazar_filter
from nova import objects
from nova import test
from nova.tests.unit.scheduler import fakes
from nova.virt import fake
from oslo_config import cfg

FLAVOR_EXTRA_SPEC = "aggregate_instance_extra_specs:reservation"


class BlazarFilterTestCase(test.TestCase):
    """Filter test case.

    This test case provides tests for the schedule filters available
    on Blazar.
    """
    def setUp(self):
        super(BlazarFilterTestCase, self).setUp()

        # Let's have at hand a brand new blazar filter
        self.f = blazar_filter.BlazarFilter()

        # A fake host state
        self.host = fakes.FakeHostState('host1', 'node1', {})

        # A fake instance (which has a reservation id 'r-fakeres')
        fake.FakeInstance('instance1', 'Running', '123')

        # And a base spec_obj
        self.spec_obj = objects.RequestSpec(
            project_id='fakepj',
            scheduler_hints={},
            flavor=objects.Flavor(flavorid='flavor-id1', extra_specs={})
        )

    def test_blazar_filter_no_pool_available_requested(self):

        # Given the host doesn't belong to any pool
        self.host.aggregates = []

        # And the 'r-fakeres' pool is requested in the filter
        self.spec_obj.scheduler_hints = {'reservation': ['r-fakeres']}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall NOT pass
        self.assertFalse(self.host.passes)

    def test_blazar_filter_no_pool_available_not_requested(self):

        # Given the host doesn't belong to any pool
        self.host.aggregates = []

        # And the filter doesn't require any pool (using filter as in setup())

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall pass
        self.assertTrue(self.host.passes)

    def test_blazar_filter_host_in_freepool_and_none_requested(self):

        # Given the host is in the free pool (named "freepool")
        self.host.aggregates = [
            objects.Aggregate(
                name=cfg.CONF['blazar:physical:host'].aggregate_freepool_name,
                metadata={'availability_zone': 'unknown',
                          self.spec_obj.project_id: True})]

        # And the filter doesn't require any pool (using filter as in setup())

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall NOT pass
        self.assertFalse(self.host.passes)

    def test_blazar_filter_host_in_pool_none_requested(self):

        # Given the host belongs to the 'r-fakeres' reservation pool
        self.host.aggregates = [
            objects.Aggregate(
                name='r-fakeres',
                metadata={'availability_zone': (cfg
                                                .CONF['blazar:physical:host']
                                                .blazar_az_prefix) + 'XX',
                          self.spec_obj.project_id: True})]

        # And the filter doesn't require any pool (using filter as in setup())

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall NOT pass
        self.assertFalse(self.host.passes)

    def test_blazar_filter_host_in_another_pool(self):

        # Given the host belongs to a pool different to 'r-fakeres'
        self.host.aggregates = [
            objects.Aggregate(
                name='not_the_r-fakeres_pool',
                metadata={'availability_zone': 'unknown',
                          self.spec_obj.project_id: True})]

        # And the 'r-fakeres' pool is requested in the filter
        self.spec_obj.scheduler_hints = {'reservation': ['r-fakeres']}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall NOT pass
        self.assertFalse(self.host.passes)

    def test_blazar_filter_host_not_auth_in_current_tenant(self):

        # Given the host is NOT authorized in the current tenant
        # And thee pool name is NOT 'r-fakeres'
        self.host.aggregates = [
            objects.Aggregate(
                name='r-fackers',
                metadata={'availability_zone': 'unknown',
                          self.spec_obj.project_id: False})]

        # And the 'r-fakeres' pool is requested in the filter
        self.spec_obj.scheduler_hints = {'reservation': ['r-fakeres']}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall NOT pass
        self.assertFalse(self.host.passes)

    def test_blazar_filter_host_auth_in_current_tenant(self):

        # Given the host is authorized in the current tenant
        # And the pool name is 'r-fakeres'
        self.host.aggregates = [
            objects.Aggregate(
                name='r-fakeres',
                metadata={'availability_zone': (cfg
                                                .CONF['blazar:physical:host']
                                                .blazar_az_prefix),
                          self.spec_obj.project_id: True})]

        # And the 'r-fakeres' pool is requested in the filter
        self.spec_obj.scheduler_hints = {'reservation': ['r-fakeres']}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall pass
        self.assertTrue(self.host.passes)

    def test_blazar_filter_host_authorized_by_owner(self):
        # Given the host blazar owner is the current project id
        # And the pool name is 'r-fakeres'
        self.host.aggregates = [
            objects.Aggregate(
                name='r-fakeres',
                metadata={'availability_zone': (cfg
                                                .CONF['blazar:physical:host']
                                                .blazar_az_prefix),
                          cfg.CONF['blazar:physical:host'].blazar_owner: (
                              self.spec_obj.project_id),
                          self.spec_obj.project_id: False})]

        # And the 'r-fakeres' pool is requested in the filter
        self.spec_obj.scheduler_hints = {'reservation': ['r-fakeres']}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall pass
        self.assertTrue(self.host.passes)

    def test_blazar_filter_host_not_authorized_by_owner(self):

        # Given the host blazar owner is NOT the current project id
        # And the pool name is 'r-fakeres'
        self.host.aggregates = [
            objects.Aggregate(
                name='r-fakeres',
                metadata={'availability_zone': (cfg
                                                .CONF['blazar:physical:host']
                                                .blazar_az_prefix),
                          cfg.CONF['blazar:physical:host'].blazar_owner: (
                              'another_project_id')})]

        # And the 'r-fakeres' pool is requested in the filter
        self.spec_obj.scheduler_hints = {'reservation': ['r-fakeres']}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall NOT pass
        self.assertFalse(self.host.passes)

    def test_blazar_filter_host_not_in_requested_pools(self):

        # Given the host is in the free pool
        self.host.aggregates = [
            objects.Aggregate(
                name=cfg.CONF['blazar:physical:host'].aggregate_freepool_name,
                metadata={'availability_zone': 'unknown'})]

        # And the 'r-fakeres' pool is requested in the filter
        self.spec_obj.scheduler_hints = {'reservation': ['r-fakeres']}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall NOT pass
        self.assertFalse(self.host.passes)

    def test_blazar_filter_unicode_requested_pool(self):

        # Given the host is in a pool with unicode characters
        self.host.aggregates = [
            objects.Aggregate(
                name=U'r-fakeres',
                metadata={'availability_zone': (cfg
                                                .CONF['blazar:physical:host']
                                                .blazar_az_prefix),
                          self.spec_obj.project_id: True})]

        # And the filter is requesting for a host with the same name (ucode)
        self.spec_obj.scheduler_hints = {'reservation': [U'r-fakeres']}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(
            self.host,
            self.spec_obj)

        # Then the host shall pass
        self.assertTrue(self.host.passes)

    def test_instance_reservation_requested(self):

        # A host is not in any aggregate
        self.host.aggregates = []

        # And instance-reservation-id1 is requested by an instance
        self.spec_obj.flavor.extra_specs = {
            FLAVOR_EXTRA_SPEC: 'instance-reservation-id1'}
        self.spec_obj.flavor.flavorid = 'instance-reservation-id1'

        self.host.passes = self.f.host_passes(self.host, self.spec_obj)

        self.assertTrue(self.host.passes)

    def test_blazar_filter_host_in_freepool_for_preemptibles(self):

        # Given preemptibles are allowed
        cfg.CONF.set_override('allow_preemptibles', True,
                              group='blazar:physical:host')
        self.addCleanup(cfg.CONF.clear_override, 'allow_preemptibles',
                        group='blazar:physical:host')

        # Given the host is in the free pool
        self.host.aggregates = [
            objects.Aggregate(
                name='freepool',
                metadata={'availability_zone': ''})]

        # And the instance is launched with a flavor marked as preemptible
        self.spec_obj.flavor.extra_specs = {'blazar:preemptible': 'true'}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(self.host, self.spec_obj)

        # Then the host shall pass
        self.assertTrue(self.host.passes)

    def test_blazar_filter_host_in_preemptibles(self):

        # Given preemptibles are allowed and dedicated aggregate is used
        cfg.CONF.set_override('allow_preemptibles', True,
                              group='blazar:physical:host')
        self.addCleanup(cfg.CONF.clear_override, 'allow_preemptibles',
                        group='blazar:physical:host')
        cfg.CONF.set_override('preemptible_aggregate', 'preemptibles',
                              group='blazar:physical:host')
        self.addCleanup(cfg.CONF.clear_override, 'preemptible_aggregate',
                        group='blazar:physical:host')

        # Given the host is in the preemptibles aggregate
        self.host.aggregates = [
            objects.Aggregate(
                name='preemptibles',
                metadata={'availability_zone': ''})]

        # And the instance is launched with a flavor marked as preemptible
        self.spec_obj.flavor.extra_specs = {'blazar:preemptible': 'true'}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(self.host, self.spec_obj)

        # Then the host shall pass
        self.assertTrue(self.host.passes)

    def test_blazar_filter_host_not_in_preemptibles(self):

        # Given preemptibles are allowed and dedicated aggregate is used
        cfg.CONF.set_override('allow_preemptibles', True,
                              group='blazar:physical:host')
        self.addCleanup(cfg.CONF.clear_override, 'allow_preemptibles',
                        group='blazar:physical:host')
        cfg.CONF.set_override('preemptible_aggregate', 'preemptibles',
                              group='blazar:physical:host')
        self.addCleanup(cfg.CONF.clear_override, 'preemptible_aggregate',
                        group='blazar:physical:host')

        # Given the host is in the free pool
        self.host.aggregates = [
            objects.Aggregate(
                name=cfg.CONF['blazar:physical:host'].aggregate_freepool_name,
                metadata={'availability_zone': 'unknown',
                          self.spec_obj.project_id: True})]

        # And the instance is launched with a flavor marked as preemptible
        self.spec_obj.flavor.extra_specs = {'blazar:preemptible': 'true'}

        # When the host goes through the filter
        self.host.passes = self.f.host_passes(self.host, self.spec_obj)

        # Then the host shall NOT pass
        self.assertFalse(self.host.passes)



# Copyright (c) 2013 Bull.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from blazarnova.i18n import _

from nova.scheduler import filters
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils.strutils import bool_from_string

LOG = logging.getLogger(__name__)

FLAVOR_EXTRA_SPEC = "aggregate_instance_extra_specs:reservation"
FLAVOR_PREEMPTIBLE = "blazar:preemptible"

opts = [
    cfg.StrOpt('aggregate_freepool_name',
               default='freepool',
               help='Name of the special aggregate where all hosts '
                    'are candidate for physical host reservation'),
    cfg.BoolOpt('allow_preemptibles',
                default=False,
                help='Whether to allow preemptible instances to be scheduled '
                     'on hosts in the preemptible aggregate'),
    cfg.StrOpt('preemptible_aggregate',
               default='freepool',
               help='Name of the aggregate where hosts can run preemptible '
                    'instances'),
    cfg.StrOpt('project_id_key',
               default='blazar:tenant',
               help='Aggregate metadata value for key matching project_id'),
    cfg.StrOpt('blazar_owner',
               default='blazar:owner',
               help='Aggregate metadata key for knowing owner project_id'),
    cfg.StrOpt('blazar_az_prefix',
               default='blazar_',
               help='Prefix for Availability Zones created by Blazar')
]

cfg.CONF.register_opts(opts, 'blazar:physical:host')


class BlazarFilter(filters.BaseHostFilter):
    """Blazar Filter for nova-scheduler."""

    run_filter_once_per_request = True

    def fetch_blazar_pools(self, host_state):
        # Get any reservation pools this host is part of
        # Note this include possibly the freepool
        aggregates = host_state.aggregates
        pools = []
        for agg in aggregates:
            if (agg.availability_zone and
                    str(agg.availability_zone).startswith(
                        cfg.CONF['blazar:physical:host'].blazar_az_prefix)
                    # NOTE(hiro-kobayashi): following 2 lines are for keeping
                    # backward compatibility
                    or str(agg.availability_zone).startswith('blazar:')):
                pools.append(agg)

            if agg.name in [
                    cfg.CONF['blazar:physical:host'].aggregate_freepool_name,
                    cfg.CONF['blazar:physical:host'].preemptible_aggregate]:
                pools.append(agg)

        return pools

    def host_reservation_request(self, host_state, spec_obj, requested_pools):

        pools = self.fetch_blazar_pools(host_state)

        for pool in [p for p in pools if p.name in requested_pools]:
            # Check tenant is allowed to use this Pool

            # NOTE(sbauza): Currently, the key is only the project_id,
            #  but later will possibly be blazar:tenant:{project_id}
            key = spec_obj.project_id
            access = pool.metadata.get(key)
            if access:
                return True
            # NOTE(sbauza): We also need to check the blazar:owner key
            #  until we modify the reservation pool for including the
            #  project_id key as for any other extra project
            owner = cfg.CONF['blazar:physical:host'].blazar_owner
            owner_project_id = pool.metadata.get(owner)
            if owner_project_id == spec_obj.project_id:
                return True
            LOG.info(_("Unauthorized request to use Pool "
                       "%(pool_id)s by tenant %(tenant_id)s"),
                     {'pool_id': pool.name,
                      'tenant_id': spec_obj.project_id})
            return False
        return False

    def host_passes(self, host_state, spec_obj):
        """Check if a host in a pool can be used for a request

        A host is in a pool if it is a member of an aggregate that has
        a metadata item with a key value of "blazar:owner"

        If the user does not pass "reservation=<id>" as a hint then only
        hosts which are not in a pool (including freepool) pass.

        If the user does pass "reservation=<id>" as a hint then the host only
        passes if it is a member of the specified pool and that pool
        has a metadata key of either :
            - "blazar:owner=tenant_id (which grants automatically all the
                users from the tenant from which the request came from)
            - or, "tenant_id=blazar:tenant" (which grants extra tenants for
                the reservation)
        """

        # Find which Pools the user wants to use (if any)
        requested_pools = spec_obj.get_scheduler_hint('reservation')
        if isinstance(requested_pools, str):
            requested_pools = [requested_pools]

        # the request is host reservation
        if requested_pools:
            return self.host_reservation_request(host_state, spec_obj,
                                                 requested_pools)

        extra_specs = spec_obj.flavor.extra_specs
        # the request is instance reservation
        if FLAVOR_EXTRA_SPEC in extra_specs.keys():
            # Scheduling requests for instance reservation are processed by
            # other Nova filters: AggregateInstanceExtraSpecsFilter,
            # AggregateMultiTenancyIsolation, and
            # ServerGroupAntiAffinityFilter. What BlazarFilter needs to
            # do is just pass the host if the request has an instance
            # reservation key.
            return True

        allow_preempt = cfg.CONF['blazar:physical:host'].allow_preemptibles
        preemptible = bool_from_string(
            extra_specs.get(FLAVOR_PREEMPTIBLE, False))
        blazar_pools = self.fetch_blazar_pools(host_state)
        # If the request is for a preemptible instance and they are allowed
        if allow_preempt and preemptible:
            if (len(blazar_pools) == 1 and blazar_pools[0].name ==
                    cfg.CONF['blazar:physical:host'].preemptible_aggregate):
                # Pass host if it only belongs to the preemptibles aggregate
                LOG.info("Host %s allowed for preemptibles" % host_state)
                return True
            return False

        if blazar_pools:
            # Host is in a blazar pool and non reservation request
            LOG.info("Host is in a reservation aggregate or in the freepool")
            return False

        return True
